// Copyright 2011-2012 Daniele Panozzo, panozzo@inf.ethz.ch. All rights reserved.
// The code in this file cannot be used for commercial
// applications without prior written permission from Daniele Panozzo

#include "visualizer.h"

#include <iomanip>

#include <QtGui/QImage>
#include <QtCore/QString>
#include <QtOpenGL/QGLWidget>

#include <QtCore/QFile>

#include "state.h"

#include <AntTweakBar.h>

using namespace Eigen;

Visualizer::Visualizer (int w, int h)
{
    width = w;
    height = h;
	
    meshEnabled = false;
    textureEnabled = true;
    saliencyEnabled = true;

    meshEnabled2 = false;
    textureEnabled2 = true;
    saliencyEnabled2 = false;

    isPainting = false;
}

void Visualizer::initMesh ()
{

    State* st = State::Instance();
    st->init();
    
	resetView();
    updateGL ();
	
	gridColor[0] = gridColor[1] = gridColor[2] = 0.7;
	
}

void Visualizer::updateGL ()
{
    glutPostRedisplay ();
}

void Visualizer::initializeGL ()
{
	glDisable(GL_CULL_FACE);
    glShadeModel(GL_SMOOTH);
	
    GLenum err = glewInit();
    if (err != GLEW_OK )
        assert(0);
}

void Visualizer::resizeGL (int w, int h)
{
    width = w;
    height = h;
    glViewport (0, 0, (GLsizei) w, (GLsizei) h);
    initializeGL ();
}

void Visualizer::paintCropBorder()
{
    State* st = State::Instance();
    glLineWidth(3);
    glColor3f(153.0/255,204.0/255,51.0/255);
    glBegin(GL_QUADS);
    glVertex2d(st->cropBL[0], st->cropBL[1]);
    glVertex2d(st->cropTR[0], st->cropBL[1]);
    glVertex2d(st->cropTR[0], st->cropTR[1]);
    glVertex2d(st->cropBL[0], st->cropTR[1]);
    glEnd();
}


void Visualizer::paintGL_aux(VectorXd& x, VectorXd& y, VectorXd& u, VectorXd& v, bool textureEnabled, bool saliencyEnabled, bool meshEnabled, bool disableMipmap, bool enableTextureCrop)
{
    State* st = State::Instance();
    // Render image
    if (textureEnabled)
    {
		glEnable( GL_TEXTURE_2D );
		glDisable(GL_LIGHTING);
        glColor3f(1,1,1);
		glEnable(GL_POLYGON_OFFSET_FILL);
		glPolygonOffset(10.0,1);
        
        if (!disableMipmap)
            glBindTexture( GL_TEXTURE_2D, st->textureImage );
        else
            glBindTexture( GL_TEXTURE_2D, st->textureImageLinear );
        
        glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
        
        // Crop modifiers for texture coordinates
        double ul = 0;
        double vl = 0;
        
        double um = 1;
        double vm = 1;

        if (enableTextureCrop) // if this is enabled then it is WP2 and we must rescale the texture
        {

            ul = st->cropBL[0]/st->originalSize[0];
            vl = st->cropBL[1]/st->originalSize[1];
            
            double ur = st->cropTR[0]/st->originalSize[0];
            double vr = st->cropTR[1]/st->originalSize[1];
            
            um = ur - ul;
            vm = vr - vl;
        }
        
        glBegin(GL_QUADS);
        for(int i=0; i<x.size()-1;++i)
        {
            for(int j=0; j<y.size()-1; ++j)
            {
                glTexCoord2d(ul + um * u[i],     vl + vm * v[j]);
                glVertex3f(  x[i],     y[j],0);
                glTexCoord2d(ul + um * u[i+1],   vl + vm * v[j]);
                glVertex3f(  x[i+1],   y[j],0);
                glTexCoord2d(ul + um * u[i+1],   vl + vm * v[j+1]);
                glVertex3f(  x[i+1],   y[j+1],0);
                glTexCoord2d(ul + um * u[i],     vl + vm * v[j+1]);
                glVertex3f(  x[i],     y[j+1],0);
            }
        }
        glEnd();

        glDisable(GL_POLYGON_OFFSET_FILL);
    }
    
    // Render image
    if (saliencyEnabled)
    {
        glEnable(GL_BLEND);
        //glBlendFunc (GL_ONE_MINUS_DST_ALPHA, GL_DST_ALPHA);
        glBlendFunc (GL_ONE, GL_ONE);
        
        glEnable( GL_TEXTURE_2D );
		glDisable(GL_LIGHTING);
        glColor3f(1,1,1);
		glEnable(GL_POLYGON_OFFSET_FILL);
		glPolygonOffset(10.0,1);
        
        glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
        
        glBindTexture( GL_TEXTURE_2D, st->textureSaliency );
        
        glBegin(GL_QUADS);
        for(int i=0; i<x.size()-1;++i)
        {
            for(int j=0; j<y.size()-1; ++j)
            {
                glTexCoord2d(u[i],     v[j]);
                glVertex3f(  x[i],     y[j],0);
                glTexCoord2d(u[i+1],   v[j]);
                glVertex3f(  x[i+1],   y[j],0);
                glTexCoord2d(u[i+1],   v[j+1]);
                glVertex3f(  x[i+1],   y[j+1],0);
                glTexCoord2d(u[i],     v[j+1]);
                glVertex3f(  x[i],     y[j+1],0);
            }
        }
        glEnd();

        glBindTexture( GL_TEXTURE_2D, st->textureSaliencyLoaded );
        
        glBegin(GL_QUADS);
        for(int i=0; i<x.size()-1;++i)
        {
            for(int j=0; j<y.size()-1; ++j)
            {
                glTexCoord2d(u[i],     v[j]);
                glVertex3f(  x[i],     y[j],0);
                glTexCoord2d(u[i+1],   v[j]);
                glVertex3f(  x[i+1],   y[j],0);
                glTexCoord2d(u[i+1],   v[j+1]);
                glVertex3f(  x[i+1],   y[j+1],0);
                glTexCoord2d(u[i],     v[j+1]);
                glVertex3f(  x[i],     y[j+1],0);
            }
        }
        glEnd();

        glDisable(GL_POLYGON_OFFSET_FILL);
        glDisable(GL_BLEND);
        
    }
    
    // Render grid
    if (meshEnabled)
    {
        glDisable( GL_TEXTURE_2D );
		glDisable(GL_LIGHTING);
        
        glDisable(GL_DEPTH_TEST);
		glEnable(GL_POLYGON_OFFSET_FILL);
		glPolygonOffset(10.0,1);
        
        glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
        
        glDisable( GL_TEXTURE_2D );
        glColor3f(gridColor[0],gridColor[1],gridColor[2]);
        
        // Crop modifiers for texture coordinates
        double ul = 0;
        double vl = 0;
        
        double um = 1;
        double vm = 1;
        
        if (!enableTextureCrop) // if this is disabled then it is WP1 and we must rescale the texture
        {
            ul = st->cropBL[0];
            vl = st->cropBL[1];
            
            double ur = st->cropTR[0];
            double vr = st->cropTR[1];
            
            um = (ur - ul)/st->originalSize[0];
            vm = (vr - vl)/st->originalSize[1];
        }
        
        glBegin(GL_QUADS);
        for(int i=0; i<x.size()-1;++i)
        {
            for(int j=0; j<y.size()-1; ++j)
            {
                glVertex3f(  ul + um * x[i],     vl + vm * y[j],0);
                glVertex3f(  ul + um * x[i+1],   vl + vm * y[j],0);
                glVertex3f(  ul + um * x[i+1],   vl + vm * y[j+1],0);
                glVertex3f(  ul + um * x[i],     vl + vm * y[j+1],0);
            }
        }
        glEnd();
        
        if (!enableTextureCrop)
            paintCropBorder();
        
		glEnable(GL_TEXTURE);
        glDisable(GL_POLYGON_OFFSET_FILL);
        glEnable(GL_DEPTH_TEST);
    }
    
    glEnable(GL_DEPTH_TEST);
}

void Visualizer::paintGL ()
{
    //glClearColor(223.0/255.0,223.0/255.0,223.0/255.0,255.0);
  glClearColor(255.0,255.0,255.0,255.0);
    glClear (GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	
	glDisable(GL_DEPTH_TEST);
	
    glMatrixMode (GL_PROJECTION);
    glLoadIdentity ();
	
    int wp1width = width/2;
    int wp2width = width - wp1width; // compensate for odd sizes

    // WP2 - Drawed first since we do not need the transformation matrices for unprojecting
    
    float sizew, sizeh;
	float ratio = (double)wp2width/(double)height;
	if (wp2width > height)
	{
		sizew = m_viewRadius*ratio;
		sizeh = m_viewRadius;
	}
	else
	{
		sizew = m_viewRadius;
		sizeh = m_viewRadius*((double)height/(double)wp2width);
	}
	
    glViewport(wp1width,0, wp2width, height);
    
	gluOrtho2D(m_viewCenter[0]-sizew, m_viewCenter[0]+sizew, m_viewCenter[1]-sizeh, m_viewCenter[1]+sizeh);
	
    glMatrixMode (GL_MODELVIEW);
    glLoadIdentity ();
	
	
    glPushMatrix();
	
    glLineWidth (1.0);
	glColor3f(0.0,0.0,0.0);
	
	glPointSize(5.0);
	
	glColor3f(1, 1, 1);
	
    State* st = State::Instance();
    
    glDisable(GL_DEPTH_TEST);
    
    if (st->useExtGrid)
        paintGL_aux(st->extGridX, st->extGridY, st->extGridU, st->extGridV, textureEnabled2, saliencyEnabled2, meshEnabled2,false,true);
    else
        paintGL_aux(st->gridX,st->gridY, st->gridU,st->gridV, textureEnabled2, saliencyEnabled2, meshEnabled2,false,true);
    
    glPopMatrix();

    // WP1
	ratio = (double)wp1width/(double)height;
	if (wp1width > height)
	{
		sizew = m_viewRadius*ratio;
		sizeh = m_viewRadius;
	}
	else
	{
		sizew = m_viewRadius;
		sizeh = m_viewRadius*((double)height/(double)wp1width);
	}
	
    glViewport(0,0, wp1width, height);

	gluOrtho2D(m_viewCenter[0]-sizew, m_viewCenter[0]+sizew, m_viewCenter[1]-sizeh, m_viewCenter[1]+sizeh);
	
    glMatrixMode (GL_MODELVIEW);
    glLoadIdentity ();
	
	
    glPushMatrix();
	
    glLineWidth (1.0);
	glColor3f(0.0,0.0,0.0);
	
	glPointSize(5.0);
	
	glColor3f(1, 1, 1);
	
    glDisable(GL_DEPTH_TEST);
    
    paintGL_aux(st->gridXOri,st->gridYOri, st->gridU,st->gridV, textureEnabled, saliencyEnabled, meshEnabled);
    
    glPopMatrix();
	
}

void Visualizer::mousePress (int button, int state, int x, int y)
{
    State* st = State::Instance();

	Point3d center = unproject(x, height-y, 0);
 
    if (state == GLUT_DOWN) // Button pressed
    {
        
        int km = glutGetModifiers();
        
        if (km & GLUT_ACTIVE_CTRL)
        {
            isPainting = true;
            st->eraserActive = false;
            mouseMove(x,y);
        }
        else if (km & GLUT_ACTIVE_ALT)
        {
            isPainting = true;
            st->eraserActive = true;
            mouseMove(x,y);
        }
        else
        {
            if (button == GLUT_RIGHT_BUTTON)
                m_viewCenter = Vector2d(center[0],center[1]);
        }
    }
    else // Button released
    {
        if (isPainting)
        {
            isPainting = false;    
            //            State::Instance()->solve();
        }
    }

}

double brushWeight(double x)
{
    double val = exp(-pow((x*3),2.0)/2)*1.5;
    if (val > 1.0)
        val = 1.0;
    return val;
}

void Visualizer::mouseMove (int x, int y)
{
    if (isPainting)
    {
        Point3d center3d = unproject(x, height-y, 0);
        Point2d center(center3d[0],center3d[1]);
        
        State* st = State::Instance();
        
        double r   = st->brushRadius;
        double in  = st->brushIntensity;
		Point2d low  = center - Point2d(r,r);
        Point2d high = center + Point2d(r,r);
        
        for(int i=floor(low[0]); i < ceil(high[0]); ++i)
        {
            for(int j=floor(low[1]); j < ceil(high[1]); ++j)
            {
				double d = (Point2d(0.5+i,0.5+j)-center).norm();
				if(d < r) // paint in a circle
				{
					if (i >= 0 && i < st->originalSize[0] && j >= 0 && j < st->originalSize[1])
					{
						double addedSaliency = in;
						double prevSaliency = st->getSal(i,j);
						double smoothnessFactor = 1.0;
						double sign = 1.0;

						if(st->brushSmooth)
						{
							smoothnessFactor = brushWeight(d/r);
						}
						if(st->eraserActive)
						{
							sign = -1.0;
						}

						double newSaliency = prevSaliency + sign*smoothnessFactor*addedSaliency;
						newSaliency = newSaliency > 255.0 ? 255.0 : newSaliency;
						newSaliency = newSaliency < 1.0 ? 1.0 : newSaliency;
						st->setSal(i, j, newSaliency);
					}
				}
            }
        }
        st->updateSaliencyLocal();
        st->solve();
    }
}

void Visualizer::keyPress(unsigned char k)
{
    State* st = State::Instance();

	if (k == 'z')
		m_viewRadius *= 1.2;
	if (k == 'x')
		m_viewRadius *= 0.8;
    if (k == 'u')
        st->solveUniform(); 
    if (k == 'i')
        st->solve();
	updateGL();
}

void Visualizer::setView(const Point2d& center, float radius)
{
	m_viewCenter = center;
	m_viewRadius = radius;
}

void Visualizer::resetView()
{
    State* state = State::Instance();
    Vector2d center;
    center[0] = (state->gridXOri(0) + state->gridXOri(state->gridXOri.size()-1)/2);
    center[1] = (state->gridYOri(0) + state->gridYOri(state->gridYOri.size()-1))/2;

    Vector2d size;
    size[0] = (-state->gridXOri(0) + state->gridXOri(state->gridXOri.size()-1))/2;
    size[1] = (-state->gridYOri(0) + state->gridYOri(state->gridYOri.size()-1))/2;
    
	setView(center,max(size[0],size[1])*1.1);
}

void Visualizer::renderText(Point3d p, string str, bool _small) 
{  
	const char *c = str.c_str();
	glRasterPos3f(p[0], p[1], p[2]);
	for (; *c != '\0'; c++) {
		glutBitmapCharacter(_small ? GLUT_BITMAP_TIMES_ROMAN_10 : GLUT_BITMAP_TIMES_ROMAN_24, *c);
	}
}

string Visualizer::convertInt(int number)
{
	stringstream ss;//create a stringstream
	ss << number;//add number to the stream
	return ss.str();//return a string with the contents of the stream
}

string Visualizer::convertDouble(double number)
{
	stringstream ss;//create a stringstream
	ss << setprecision(4) << number;//add number to the stream
	return ss.str();//return a string with the contents of the stream
}

Point3d Visualizer::project(const Point3d& p)
{
	double model_view_matrix[16];
	double projection_matrix[16];
	int    view_port[4];
	glGetDoublev(GL_MODELVIEW_MATRIX,  model_view_matrix);
	glGetDoublev(GL_PROJECTION_MATRIX, projection_matrix);
	glGetIntegerv(GL_VIEWPORT, view_port);
	double x, y, z;
	gluProject(p[0], p[1], p[2], model_view_matrix, projection_matrix,
			   view_port, &x, &y, &z);
	return Point3d(x, y, z);
}

Point3d Visualizer::unproject(double sx, double sy, double sz)
{
	double model_view_matrix[16];
	double projection_matrix[16];
	int    view_port[4];
	glGetDoublev(GL_MODELVIEW_MATRIX,  model_view_matrix);
	glGetDoublev(GL_PROJECTION_MATRIX, projection_matrix);
	glGetIntegerv(GL_VIEWPORT, view_port);
	double x, y, z;
	gluUnProject(sx, sy, sz, model_view_matrix, projection_matrix,
				 view_port, &x, &y, &z);
	return Point3d(x, y, z);
}

void Visualizer::renderToFile(const char* str)
{
    
#define CHUNKSIZE 100 

    
    State* st = State::Instance();

    double w = st->currentSize(0);
    double h = st->currentSize(1);
    
    double xChunk = ceil(w/(float)CHUNKSIZE);
    double yChunk = ceil(h/(float)CHUNKSIZE);

    QImage image(w, h, QImage::Format_RGB32);
    
    for (int x=0; x < xChunk; ++x)
    {
        for (int y=0; y < yChunk; ++y)
        {
            glClearColor (0,0,0,0);
            glClear (GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
            
            glMatrixMode(GL_PROJECTION);
            glLoadIdentity();

            glViewport(0,0, CHUNKSIZE, CHUNKSIZE);

            // One chunk
            double l = x * CHUNKSIZE;
            double b = y * CHUNKSIZE;
            double r = l + CHUNKSIZE;
            double t = b + CHUNKSIZE;
            
            gluOrtho2D(l,r,b,t);

            glMatrixMode(GL_MODELVIEW);
            glLoadIdentity();
            
            if (st->useExtGrid)
                paintGL_aux(st->extGridX, st->extGridY, st->extGridU, st->extGridV, textureEnabled2, saliencyEnabled2, meshEnabled2,true,true);
            else
                paintGL_aux(st->gridX,st->gridY, st->gridU,st->gridV, textureEnabled2, saliencyEnabled2, meshEnabled2,true,true);
            
            unsigned char* buff = (unsigned char*)calloc(CHUNKSIZE*CHUNKSIZE*3,sizeof(unsigned char));
            
            glReadPixels(0,
                         0,
                         CHUNKSIZE,
                         CHUNKSIZE,
                         GL_RGB,
                         GL_UNSIGNED_BYTE,
                         buff
                         );

            
            for (int i=0; i<CHUNKSIZE; ++i)
            {
                for (int j=0; j<CHUNKSIZE; ++j)
                {
                    QRgb value;
                    unsigned char r,g,b;
                    
                    r = buff[3*(i*CHUNKSIZE+j)+0];
                    g = buff[3*(i*CHUNKSIZE+j)+1];
                    b = buff[3*(i*CHUNKSIZE+j)+2];
                    
                    value = qRgb(r,g,b);
                    
                    // swap
                    int ic = i;
                    int jc = j;
                    std::swap(ic,jc);
                    jc = CHUNKSIZE - jc;
                    
                    int ifinal = ic + x * CHUNKSIZE;
                    int jfinal = jc + (yChunk - y - 1) * CHUNKSIZE;
                    
                    jfinal -= CHUNKSIZE - (h - (CHUNKSIZE * (yChunk-1)));
                    
                    if (ifinal >= 0 && ifinal < w && jfinal >= 0 && jfinal < h)
                    {
                        image.setPixel(ifinal, jfinal, value);
                    }
                }
            }
            
            free(buff);
        }
    }

    image.save(QString(str));
    
}

void Visualizer::saveSnapshot(const char* str)
{
    int w = width;
    int h = height;
    
    QImage image(w, h, QImage::Format_RGB32);

    paintGL();
    TwDraw ();

    unsigned char* buff = (unsigned char*)calloc(w*h*3,sizeof(unsigned char));
            
    glReadPixels(0,
                 0,
                 w,
                 h,
                 GL_RGB,
                 GL_UNSIGNED_BYTE,
                 buff
                 );
            
            
    for (int i=0; i<w; ++i)
    {
        for (int j=0; j<h; ++j)
        {
            QRgb value;
            unsigned char r,g,b;

            int i0 = 3*(j*w+i)+0;
            int i1 = 3*(j*w+i)+1;
            int i2 = 3*(j*w+i)+2;
            r = buff[i0];
            g = buff[i1];
            b = buff[i2];
            
            value = qRgb(r,g,b);
            
            image.setPixel(i,h-j-1, value);
                    
        }
    }
            
    free(buff);
    
    image.save(QString(str) + ".png");
    
}

